//         TDVFontsCombo & TDVFontsList 1.1
//              (c) Jakub Dusek (1999)
//
//  Tested in Delphi 2 and 4, but I think that allready
//       works in other 32 bit Delphi versions
//         (mail me some possible problems)
//
// This source code is not public domain and can
//     be distributed only with this header
//           and included readme file
//
//     For updates and more components visit
//          http://drivi.misto.cz/delphi
//
//      Mail me some bugs and good ideas to
//                 drivi@seznam.cz

{$IFNDEF VER80}
 {$IFNDEF VER90}
  {$IFNDEF VER93}
    {$IFNDEF VER100}
      {$IFNDEF VER110}
        {$DEFINE DV4}
      {$ENDIF}
    {$ENDIF}
  {$ENDIF}
 {$ENDIF}
{$ENDIF}

unit DVFonts;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls,
  Forms,Dialogs,StdCtrls,Printers;

type
  TFontType = (TrueTypeFonts, RasterFonts, DeviceFonts, OtherFonts);

  TViewFontsType = set of TFontType;

  TDVFontsCombo = class(TCustomComboBox)
  private
    FAutoUpdate:Boolean;
    FTrueTypeBmp:TBitmap;
    FRasterBmp:TBitmap;
    FDeviceBmp:TBitmap;
    FFontName:TFontName;
    FViewFontsType:TViewFontsType;
    FBmpWidth:Integer;
    procedure SetFontName(Value:TFontName);
    procedure SetViewFontsType(Value:TViewFontsType);
    procedure UpdateHeight;
    procedure CMFontChanged(var Message: TMessage);message CM_FONTCHANGED;
    procedure WMFontChange(var Message: TMessage); message WM_FONTCHANGE;
  protected
    procedure DropDown;override;
    procedure DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);override;
    procedure Change;override;
  public
    constructor Create(AOwner:TComponent);override;
    procedure CreateWnd; override;
    procedure Loaded; override;
    destructor Destroy;override;
    procedure SetTrueTypeBmp(Value:TBitmap);
    procedure SetRasterBmp(Value:TBitmap);
    procedure SetDeviceBmp(Value:TBitmap);
  published
    property AutoUpdate:Boolean read FAutoUpdate write FAutoUpdate default true;
    property FontName:TFontName read FFontName write SetFontName;
    property ViewFontsType:TViewFontsType read FViewFontsType write SetViewFontsType default [TrueTypeFonts, RasterFonts, DeviceFonts, OtherFonts];
    property TrueTypeBmp:TBitmap read FTrueTypeBmp write SetTrueTypeBmp;
    property RasterBmp:TBitmap read FRasterBmp write SetRasterBmp;
    property DeviceBmp:TBitmap read FDeviceBmp write SetDeviceBmp;
    {$IFDEF DV4}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
    property OnEndDock;
    property OnStartDock;
    {$ENDIF}
    {$IFNDEF VER90}
    property ImeMode;
    property ImeName;
    {$ENDIF}
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property DropDownCount;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDropDown;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnStartDrag;
    procedure UpdateFonts;
  end;

TDVFontsList = class(TCustomListBox)
  private
    FTrueTypeBmp:TBitmap;
    FRasterBmp:TBitmap;
    FDeviceBmp:TBitmap;
    FFontName:TFontName;
    FViewFontsType:TViewFontsType;
    FBmpWidth:Integer;
    procedure SetFontName(Value:TFontName);
    procedure SetViewFontsType(Value:TViewFontsType);
    procedure UpdateHeight;
    procedure CMFontChanged(var Message: TMessage);message CM_FONTCHANGED;
  protected
    procedure DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);override;
    procedure Click;override;
  public
    constructor Create(AOwner:TComponent);override;
    procedure CreateWnd; override;
    procedure Loaded; override;
    destructor Destroy;override;
    procedure SetTrueTypeBmp(Value:TBitmap);
    procedure SetRasterBmp(Value:TBitmap);
    procedure SetDeviceBmp(Value:TBitmap);
  published
    property FontName:TFontName read FFontName write SetFontName;
    property ViewFontsType:TViewFontsType read FViewFontsType write SetViewFontsType default [TrueTypeFonts, RasterFonts, DeviceFonts, OtherFonts];
    property TrueTypeBmp:TBitmap read FTrueTypeBmp write SetTrueTypeBmp;
    property RasterBmp:TBitmap read FRasterBmp write SetRasterBmp;
    property DeviceBmp:TBitmap read FDeviceBmp write SetDeviceBmp;
    property Align;
    {$IFDEF DV4}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
    property OnEndDock;
    property OnStartDock;
    {$ENDIF}
    {$IFNDEF VER90}
    property ImeMode;
    property ImeName;
    {$ENDIF}
    property BorderStyle;
    property Color;
    property Columns;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property ExtendedSelect;
    property Font;
    property IntegralHeight;
    property MultiSelect;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnStartDrag;
    procedure UpdateFonts;
  end;

procedure Register;

{$R DVFonts.RES}

implementation

procedure Register;
begin
  RegisterComponents('Drivi Lib', [TDVFontsCombo]);
  RegisterComponents('Drivi Lib', [TDVFontsList]);
end;

function GetItemHeight(Font: TFont): Integer;
var
  DC: HDC;
  Fnt: HFont;
  Metrics: TTextMetric;
begin
  DC := GetDC(0);
  try
    Fnt := SelectObject(DC, Font.Handle);
    GetTextMetrics(DC, Metrics);
    SelectObject(DC, Fnt);
  finally
    ReleaseDC(0, DC);
  end;
  Result := Metrics.tmHeight + 1;
end;

constructor TDVFontsCombo.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FTrueTypeBmp:=TBitmap.Create;
  FRasterBmp:=TBitmap.Create;
  FDeviceBmp:=TBitmap.Create;
  FTrueTypeBmp.Handle:=LoadBitmap(HInstance,'DVTTF');
  FRasterBmp.Handle:=LoadBitmap(HInstance,'DVFON');
  FDeviceBmp.Handle:=LoadBitmap(HInstance,'DVPRN');
  Style:=csOwnerDrawFixed;
  Sorted:=True;
  FViewFontsType:=[TrueTypeFonts, RasterFonts, DeviceFonts,OtherFonts];
  FAutoUpdate:=True;
end;

procedure TDVFontsCombo.CreateWnd;
begin
  inherited CreateWnd;
  UpdateFonts;
end;

procedure TDVFontsCombo.Loaded;
begin
  inherited Loaded;
  UpdateHeight;
  SetFontName(FFontName);
end;

destructor TDVFontsCombo.Destroy;
begin
  FTrueTypeBmp.Free;
  FRasterBmp.Free;
  FDeviceBmp.Free;
  inherited Destroy;
end;

procedure TDVFontsCombo.SetTrueTypeBmp(Value:TBitmap);
begin
  FTrueTypeBmp.Assign(Value);
  UpdateHeight;
end;

procedure TDVFontsCombo.SetRasterBmp(Value:TBitmap);
begin
  FRasterBmp.Assign(Value);
  UpdateHeight;
end;

procedure TDVFontsCombo.SetDeviceBmp(Value:TBitmap);
begin
  FDeviceBmp.Assign(Value);
  UpdateHeight;
end;

procedure TDVFontsCombo.UpdateHeight;
var i:integer;
begin
  i:=GetItemHeight(Font);
  if FTrueTypeBmp.Height>i then i:=FTrueTypeBmp.Height;
  if FRasterBmp.Height>i then i:=FRasterBmp.Height;
  if FDeviceBmp.Height>i then i:=FDeviceBmp.Height;
  itemheight:=i+2;
  Height:=itemheight;
  FBmpWidth:=FTrueTypeBmp.Width;
  if FRasterBmp.Width>FBmpWidth then FBmpWidth:=FRasterBmp.Width;
  if FDeviceBmp.Width>FBmpWidth then FBmpWidth:=FDeviceBmp.Width;
  RecreateWnd;
  Invalidate;
end;

procedure TDVFontsCombo.WMFontChange(var Message: TMessage);
begin
  if FAutoUpdate then UpdateFonts;
  inherited;
end;

procedure TDVFontsCombo.CMFontChanged(var Message: TMessage);
var i:integer;
begin
  inherited;
  UpdateHeight;
end;

procedure TDVFontsCombo.DropDown;
begin
  if FAutoUpdate then UpdateFonts;
  inherited dropdown;
end;

procedure TDVFontsCombo.Change;
begin
  FFontName:=Items[ItemIndex];
  inherited Change;
end;

procedure TDVFontsCombo.SetFontName(Value:TFontName);
begin
  if Value='' then FFontName:=Value;
  if Items.IndexOf(Value)<>-1 then
  begin
    ItemIndex:=Items.IndexOf(Value);
    FFontName:=Value;
  end;
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TDVFontsCombo.SetViewFontsType(Value:TViewFontsType);
begin
  FViewFontsType:=Value;
  UpdateFonts;
end;

procedure TDVFontsCombo.UpdateFonts;

function IsValid(Combo: TDVFontsCombo;FontType:Integer): Boolean;
begin
  Result:=False;
  if (FontType and TRUETYPE_FONTTYPE) <> 0 then
  begin
    Result:=TrueTypeFonts in Combo.ViewFontsType;
  end
  else
    if (FontType and DEVICE_FONTTYPE) <> 0 then
    begin
      Result:=DeviceFonts in Combo.ViewFontsType;
    end
    else
      if (FontType and RASTER_FONTTYPE) <> 0 then
      begin
        Result:=RasterFonts in Combo.ViewFontsType;
      end
      else
        Result:=OtherFonts in Combo.ViewFontsType;
end;

function EnumFontsProc(var EnumLogFont: TEnumLogFont;
  var TextMetric: TNewTextMetric; FontType: Integer; Data: LPARAM): Integer;
  export; stdcall;
var
  FaceName: string;
begin
  FaceName := StrPas(EnumLogFont.elfLogFont.lfFaceName);
  with TDVFontsCombo(Data) do
    if (Items.IndexOf(FaceName)=-1) and (IsValid(TDVFontsCombo(Data),FontType)) then
    begin
      Items.AddObject(FaceName, TObject(FontType));
    end;
  Result := 1;
end;

var
  DC: HDC;
  Old:string;
begin
  if Items.Count<>0 then Old:=Items[ItemIndex];
  Items.Clear;
  DC := GetDC(0);
  try
    EnumFontFamilies(DC, nil, @EnumFontsProc, Longint(Self));
    EnumFontFamilies(Printer.Handle, nil, @EnumFontsProc, Longint(Self));
  finally
    ReleaseDC(0, DC);
  end;
  if Items.indexof(Old)<>-1 then ItemIndex:=Items.indexof(Old)
  else
  begin
    if Items.Count<>0 then
    begin
      ItemIndex:=0;
      SetFontName(Items[ItemIndex]);
    end;
  end;
  if Items.Count<>0 then SetFontName(Items[ItemIndex])
  else
    SetFontName('');
end;

procedure TDVFontsCombo.DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);
var b:TBitmap;
begin
  b:=TBitmap.Create;
  Canvas.FillRect(rect);
  if (Integer(Items.Objects[Index]) and TRUETYPE_FONTTYPE) <> 0 then
    b.assign(FTrueTypeBmp)
  else
    if (Integer(Items.Objects[Index]) and DEVICE_FONTTYPE) <> 0 then
      b.assign(FDeviceBmp)
    else
      if (Integer(Items.Objects[Index]) and RASTER_FONTTYPE) <> 0 then
        b.assign(FRasterBmp);
  Canvas.BrushCopy(Bounds(Rect.Left + 2, (Rect.Top + Rect.Bottom - b.Height)
                   div 2, b.Width, b.Height), b, Bounds(0, 0, b.Width,
                   b.Height), b.TransparentColor);
  rect.Left:=rect.Left+FBmpWidth+4;
  DrawText(Canvas.Handle, PChar(Items[Index]), Length(Items[Index]),
           Rect, DT_SINGLELINE or DT_VCENTER or DT_NOPREFIX);
  b.Free;
end;

constructor TDVFontsList.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FTrueTypeBmp:=TBitmap.Create;
  FRasterBmp:=TBitmap.Create;
  FDeviceBmp:=TBitmap.Create;
  FTrueTypeBmp.Handle:=LoadBitmap(HInstance,'DVTTF');
  FRasterBmp.Handle:=LoadBitmap(HInstance,'DVFON');
  FDeviceBmp.Handle:=LoadBitmap(HInstance,'DVPRN');
  Style:=lbOwnerDrawFixed;
  Sorted:=True;
  FViewFontsType:=[TrueTypeFonts, RasterFonts, DeviceFonts,OtherFonts];
end;

procedure TDVFontsList.CreateWnd;
begin
  inherited CreateWnd;
  UpdateFonts;
end;

procedure TDVFontsList.Loaded;
begin
  inherited Loaded;
  UpdateHeight;
  SetFontName(FFontName);
end;

destructor TDVFontsList.Destroy;
begin
  FTrueTypeBmp.Free;
  FRasterBmp.Free;
  FDeviceBmp.Free;
  inherited Destroy;
end;

procedure TDVFontsList.SetTrueTypeBmp(Value:TBitmap);
begin
  FTrueTypeBmp.Assign(Value);
  UpdateHeight;
end;

procedure TDVFontsList.SetRasterBmp(Value:TBitmap);
begin
  FRasterBmp.Assign(Value);
  UpdateHeight;
end;

procedure TDVFontsList.SetDeviceBmp(Value:TBitmap);
begin
  FDeviceBmp.Assign(Value);
  UpdateHeight;
end;

procedure TDVFontsList.UpdateHeight;
var i:integer;
begin
  i:=GetItemHeight(Font);
  if FTrueTypeBmp.Height>i then i:=FTrueTypeBmp.Height;
  if FRasterBmp.Height>i then i:=FRasterBmp.Height;
  if FDeviceBmp.Height>i then i:=FDeviceBmp.Height;
  itemheight:=i+2;
  FBmpWidth:=FTrueTypeBmp.Width;
  if FRasterBmp.Width>FBmpWidth then FBmpWidth:=FRasterBmp.Width;
  if FDeviceBmp.Width>FBmpWidth then FBmpWidth:=FDeviceBmp.Width;
  RecreateWnd;
  Invalidate;
end;

procedure TDVFontsList.CMFontChanged(var Message: TMessage);
var i:integer;
begin
  inherited;
  UpdateHeight;
end;

procedure TDVFontsList.Click;
begin
  FFontName:=Items[ItemIndex];
  inherited Click;
end;

procedure TDVFontsList.SetFontName(Value:TFontName);
begin
  if Value='' then FFontName:=Value;
  if Items.IndexOf(Value)<>-1 then
  begin
    ItemIndex:=Items.IndexOf(Value);
    FFontName:=Value;
  end;
  if Assigned(OnClick) then OnClick(Self);
end;

procedure TDVFontsList.SetViewFontsType(Value:TViewFontsType);
begin
  FViewFontsType:=Value;
  UpdateFonts;
end;

procedure TDVFontsList.UpdateFonts;

function IsValid(Combo: TDVFontsList;FontType:Integer): Boolean;
begin
  Result:=False;
  if (FontType and TRUETYPE_FONTTYPE) <> 0 then
  begin
    Result:=TrueTypeFonts in Combo.ViewFontsType;
  end
  else
    if (FontType and DEVICE_FONTTYPE) <> 0 then
    begin
      Result:=DeviceFonts in Combo.ViewFontsType;
    end
    else
      if (FontType and RASTER_FONTTYPE) <> 0 then
      begin
        Result:=RasterFonts in Combo.ViewFontsType;
      end
      else
        Result:=OtherFonts in Combo.ViewFontsType;
end;

function EnumFontsProc(var EnumLogFont: TEnumLogFont;
  var TextMetric: TNewTextMetric; FontType: Integer; Data: LPARAM): Integer;
  export; stdcall;
var
  FaceName: string;
begin
  FaceName := StrPas(EnumLogFont.elfLogFont.lfFaceName);
  with TDVFontsList(Data) do
    if (Items.IndexOf(FaceName)=-1) and (IsValid(TDVFontsList(Data),FontType)) then
    begin
      Items.AddObject(FaceName, TObject(FontType));
    end;
  Result := 1;
end;

var
  DC: HDC;
  Old:string;
begin
  if Items.Count<>0 then Old:=Items[ItemIndex];
  Items.Clear;
  DC := GetDC(0);
  try
    EnumFontFamilies(DC, nil, @EnumFontsProc, Longint(Self));
    EnumFontFamilies(Printer.Handle, nil, @EnumFontsProc, Longint(Self));
  finally
    ReleaseDC(0, DC);
  end;
  if Items.indexof(Old)<>-1 then ItemIndex:=Items.indexof(Old)
  else
  begin
    if Items.Count<>0 then
    begin
      ItemIndex:=0;
      SetFontName(Items[ItemIndex]);
    end;
  end;
  if Items.Count<>0 then SetFontName(Items[ItemIndex])
  else
    SetFontName('');
end;

procedure TDVFontsList.DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);
var b:TBitmap;
begin
  b:=TBitmap.Create;
  Canvas.FillRect(rect);
  if (Integer(Items.Objects[Index]) and TRUETYPE_FONTTYPE) <> 0 then
    b.assign(FTrueTypeBmp)
  else
    if (Integer(Items.Objects[Index]) and DEVICE_FONTTYPE) <> 0 then
      b.assign(FDeviceBmp)
    else
      if (Integer(Items.Objects[Index]) and RASTER_FONTTYPE) <> 0 then
        b.assign(FRasterBmp);
  Canvas.BrushCopy(Bounds(Rect.Left + 2, (Rect.Top + Rect.Bottom - b.Height)
                   div 2, b.Width, b.Height), b, Bounds(0, 0, b.Width,
                   b.Height), b.TransparentColor);
  rect.Left:=rect.Left+FBmpWidth+4;
  DrawText(Canvas.Handle, PChar(Items[Index]), Length(Items[Index]),
           Rect, DT_SINGLELINE or DT_VCENTER or DT_NOPREFIX);
  b.Free;
end;

end.
